/* CapiServer.cpp

Implements a server for exporting ivr functionality to other hosts.

Part of the caiviar package. 

Copyright (c) 2002 MobileX AG, http://www.mobilexag.de
Copyright (c) 2002 Matthias Kramm <kramm@quiss.org>

This file is distributed under the GPL, see file COPYING for details. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../config.h"

#if defined(LINUX) || defined(CYGWIN)
#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#else
#ifndef __windows_h_included__
#define __windows_h_included__
#include <windows.h>
#endif
#endif

#include "TextToSpeech.h"

#include "ServerSocket.h"
#include "ClientSocket.h"
#include "../src/SimpleCapi.h"
#include "../src/os.h"
#include "string.h"

static ServerSocket* server = 0;
static Logger*log = 0;
static SimpleCapi*capi = 0;

static int port = 7347;
static char* realspeak_dictionary = 0;

#ifdef LINUX
static int dofaxconversion(char*filename, char*filename_sff)
{
    char buf[1024];
    sprintf(buf, "html2ps %s > tmp.ps", filename);
    printf("%s\n", buf);
    system(buf);
    sprintf(buf, "echo quit | gs -sDEVICE=cfax -sOutputFile=%s tmp.ps", filename_sff);
    printf("%s\n", buf);
    system(buf);
    return 1;
}
#else
static int dofaxconversion(char*filename, char*filename_sff)
{
    char buf[1024];
    sprintf(buf, "copy %s %s", filename, filename_sff);
    system(buf);
    return 1;
}
#endif

static char * getnextstring(char*buffer, int&pos)
{
  int oldpos = pos;
  if(!buffer[pos])
    return 0;
  char*chr = strchr(&buffer[pos],' ');
  if(chr)
  {
    pos = chr-buffer;
    while(buffer[pos]==' ' || buffer[pos]=='\t')
    {
      buffer[pos]=0;
      pos++;
    }
    return &buffer[oldpos];
  }
  else
  {
    return &buffer[oldpos];
  }
}

static Thread thread;
THREAD_FUNCTION callHandler(void*data)
{
  TextToSpeech* tts = TextToSpeech::getImplementation(log);
  tts->setParams("realspeak.dictionary", realspeak_dictionary);
  
  char buffer[8192];
  char buffer2[8192];
  ClientSocket*socket = (ClientSocket*)data;
  ISDNEvent events[256];
  int eventcount=0;
  int line;
  int ret;
  int backoff = 1;
  do {
      line = capi->open(0); // 0 = get msisdn from properties
      if(line>=0)
	  break;
      log->logf("<warning> Couldn't create a line, trying again in %d seconds\n", backoff);
      Sleep(backoff*1000);
      backoff*=2;
      if(backoff>=300) {
	  backoff=300;
      }
  } while(line<0);

  int lasteventid = 0;
  log->logf("<notice> [line %d] opened line %d\n", line, line);

  ret = socket->myprintf("open=%d\n", line);

  if(!ret)
  while(1)
  {
      socket->readLine(buffer, 8192);
      memcpy(buffer2,buffer, 8192);
      log->logf("<verbose> [line %d] socket: >%s\n", line, buffer);
      int bufpos = 0;
      char* string = getnextstring(buffer, bufpos);
      if(!string)
	break;
      if(!strcasecmp(string, "accept"))
      {
	    capi->acceptCalls(line);
	    if(socket->print("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "set"))
      {
	    char*param = getnextstring(buffer, bufpos);
	    if(!strcmp(param, "msisdn")) {
		char* msisdn = getnextstring(buffer, bufpos); 
		if(!msisdn) {
		    if(socket->print("error: missing msisdn\n"))
			break;
		    log->logf("<verbose> [line %d] socket: error: missing msisdn\n", line);
		    continue;
		}
		capi->close(line);
		line = capi->open(msisdn);
	    } else {
		socket->myprintf("error: unknown parameter: %s\n", param);
		log->logf("<verbose> [line %d] socket: error: unknown parameter: %s\n", line, param);
		continue;
	    }
	    /* the line may have changed, that's why we pass it back in the return
	       value */
	    if(socket->myprintf("ok=%d\n", line))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line, line);
      }
      else if(!strcasecmp(string, "hangup"))
      {
	    capi->hangup(line);
	    if(socket->print("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "hold"))
      {
	    int ret = capi->hold(line);
	    if(socket->myprintf("ok=%d\n",ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "retrieve"))
      {
	    int ret = capi->retrieve(line);
	    if(socket->myprintf("ok=%d\n",ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "supplemental"))
      {
	    char*string = getnextstring(buffer, bufpos);
	    if(!string) 
	      break;
	    int ret = capi->supplemental(line, string);
	    if(socket->myprintf("ok=%d\n",ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "call"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    char*timeout= getnextstring(buffer, bufpos);
	    if(!timeout)
	      break;
	    int ret = capi->call(line, msisdn, atoi(timeout));
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "call_nonblock"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    int ret = capi->call_nonblock(line, msisdn);
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "play"))
      {
	    char*filename = getnextstring(buffer, bufpos);
	    if(!filename) 
	      break;
	    int ret = capi->play(line, filename);
	    if(socket->myprintf("ok=%d\n",ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "record"))
      {
	    char*filename = getnextstring(buffer, bufpos);
	    if(!filename) 
	      break;
	    char*length = getnextstring(buffer, bufpos);
	    if(!length) 
	      break;
	    int ret = capi->recordWAV(line, filename,atoi(length)*8000);
	    if(socket->myprintf("ok=%d\n",ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "beep"))
      {
	    char*beeps = getnextstring(buffer, bufpos);
	    if(!beeps) 
	      break;
	    capi->beep(line, beeps);
	    if(socket->print("ok\n"))
		break;
      }
      else if(!strcasecmp(string, "speak"))
      {
	    char*chr1 = strchr((char*)buffer2, '\"');
	    if(!chr1)
		break;
	    chr1++;
	    char*chr2 = strchr(chr1, '\"');
	    if(!chr2)
		break;
	    *chr2=0;
	    unsigned char*data;
	    unsigned long size = 0;
	    tts->text2Stream(chr1, &data, &size);
	    if(size==0) {
		if(socket->myprintf("error=\"tts failed\"\n"))
		    break;
		log->logf("<verbose> [line %d] socket: error=\", linetts failed\"\n");
		continue;
	    }
	    int ret = capi->play(line, data, size);
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line,ret);
      }
      else if(!strcasecmp(string, "fax"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    char*filename= getnextstring(buffer, bufpos);
	    if(!filename)
	      break;
	    dofaxconversion(filename, "tmp.sff");
	    int ret = capi->fax(line, msisdn, "tmp.sff");
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line, ret);
      }
      else if(!strcasecmp(string, "stop"))
      {
	    char*jobid = getnextstring(buffer, bufpos);
	    if(!jobid) 
	      break;
	    capi->stop(line, atoi(jobid));
	    if(socket->print("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "put"))
      {
	    char*filename = getnextstring(buffer, bufpos);
	    if(!filename) 
	      break;
	    if(strchr(filename, '/') || strchr(filename, '\\') || strlen(filename)>=1020)
	    {
		if(socket->myprintf("error: no '/' or '\\' characters allowed in filename\n"))
		    break;
		log->logf("<verbose> [line %d] socket: error: no '/' or '\\' characters allowed in filename\n", line);
		continue;
	    }

	    char*sizestr= getnextstring(buffer, bufpos);
	    if(!sizestr)
	      break;
	    int size = atoi(sizestr);
	    int pos = 0;
	    char* buffer = (char*)malloc(size);
	    int ret;
	    memset(buffer, 0xaa, size);
	    if(!buffer) {
		log->logf("<verbose> [line %d] socket: error: out of memory\n", line);
		if(socket->myprintf("error: out of memory\n"))
		    break;
		continue;
	    }
	    while(pos<size)
	    {
		int len = 1024;
		if(len > size-pos)
		    len = size-pos;
		ret = socket->getBlock(&buffer[pos], len);
		if(ret<=0) break;
		pos += ret;
	    }
	    if(ret<=0) {
		break;
	    }
	    FILE*fi = fopen(filename, "wb");
	    if(!fi) {
		log->logf("<verbose> [line %d] socket: error: Couldn't open file %s\n", line, filename);
		if(socket->myprintf("error: Couldn't open %s\n", filename))
		    break;
		continue;
	    }
	    fwrite(buffer, size, 1, fi);
	    fclose(fi);
	    free(buffer);

	    if(socket->myprintf("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "del"))
      {
	    char*filename = getnextstring(buffer, bufpos);
	    if(!filename) 
	      break;
	    if(strchr(filename, '/') || strchr(filename, '\\') || strlen(filename)>=1020)
	    {
		if(socket->myprintf("error: no '/' or '\\' characters allowed in filename\n"))
		    break;
		log->logf("socket: error: no '/' or '\\' characters allowed in filename\n");
		break;
	    }

	    char buf[1024];
#ifdef LINUX
	    sprintf(buf, "rm %s", filename);
#else
	    sprintf(buf, "del %s", filename);
#endif
	    system(buf);
	    if(socket->myprintf("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "listen"))
      {
	    char*timeout = getnextstring(buffer, bufpos);
	    if(!timeout) 
		timeout = "0";
	    int t = atoi(timeout);
	    ISDNEvent event;
	    event = capi->listen(line, t);
	    /*if(t) 
		t++;
	       int i;
	       for (i = 1 ;i != t; i++) {
		event = capi->listen(line, 1);
		if(event.type != TIMEOUT)
		    break;
	    }*/
	    events[eventcount] = event;
	    if(socket->myprintf("msisdn=%s\n",event.msisdn))
		break;
	    log->logf("<verbose> [line %d] socket: msisdn=%s\n", line, event.msisdn);
	    if(socket->myprintf("called_msisdn=%s\n",event.called_msisdn))
		break;
	    log->logf("<verbose> [line %d] socket: called_msisdn=%s\n", line, event.called_msisdn);
	    if(socket->myprintf("id=%d\n",eventcount))
		break;
	    log->logf("<verbose> [line %d] socket: id=%d\n", line, eventcount);
	    if(socket->myprintf("type=%d\n", event.type))
		break;
	    log->logf("<verbose> [line %d] socket: type=%d\n", line, event.type);
	    if(socket->myprintf("jobid=%d\n", event.jobid))
		break;
	    log->logf("<verbose> [line %d] socket: jobid=%d\n", line, event.jobid);
	    if(socket->myprintf("tone=%d\n", event.tone))
		break;
	    log->logf("<verbose> [line %d] socket: tone=%d\n", line, event.tone);
	    if(socket->myprintf("reason=%d\n", event.reason))
		break;
	    log->logf("<verbose> [line %d] socket: reason=%d\n", line, event.reason);
	    if(event.type == CAPI_EVENT_INCOMING)
	        lasteventid = eventcount;
	    eventcount++;
	    eventcount&=256;
	    if(socket->myprintf("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "answer"))
      {
	    char*eventid = getnextstring(buffer, bufpos);
	    int eid = -1;
	    if(eventid && *eventid) 
	    {
		eid = atoi(eventid);
	    }
	    if(eid>=256 || eid<0)
		eid = lasteventid;
	    int ret=capi->answer(line, &events[eid]);
	    if(socket->myprintf("ok=%d\n", ret))
		break;
	    log->logf("<verbose> [line %d] socket: ok=%d\n", line, ret);
      }
      else if(!strcasecmp(string, "reject"))
      {
	    char*eventid = getnextstring(buffer, bufpos);
	    int eid = -1;
	    if(eventid && *eventid) 
	    {
		eid = atoi(eventid);
	    }
	    if(eid>=256 || eid<0)
		eid = lasteventid;
	    capi->reject(line, &events[eid], 2);
	    if(socket->myprintf("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "ignore"))
      {
	    char*eventid = getnextstring(buffer, bufpos);
	    int eid = -1;
	    if(eventid && *eventid) 
	    {
		eid = atoi(eventid);
	    }
	    if(eid>=256 || eid<0)
		eid = lasteventid;
	    capi->reject(line, &events[eid], 1);
	    if(socket->myprintf("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "forward"))
      {
	    char*msisdn = getnextstring(buffer, bufpos);
	    if(!msisdn) 
	      break;
	    int ret = capi->forward(line, msisdn);
	    if(socket->print("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "connect"))
      {
	    char*strline1 = getnextstring(buffer, bufpos);
	    if(!strline1) 
	      break;
	    char*strline2 = getnextstring(buffer, bufpos);
	    if(!strline2) 
	      break;
	    int line1,line2;
	    line1 = atoi(strline1);
	    line2 = atoi(strline2);
	    capi->connect(line1, line2);
	    if(socket->print("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
      }
      else if(!strcasecmp(string, "close"))
      {
	    if(socket->print("ok\n"))
		break;
	    log->logf("<verbose> [line %d] socket: ok\n", line);
	    break;
      }
      else 
      {
	    if(socket->print("error: Unknown command\n"))
		break;
	    log->logf("<verbose> [line %d] socket: error: Unknown command\n", line);
      }
  }
  log->logf("<verbose> [line %d] closing line %d\n", line, line);
  capi->close(line);
	 
  delete socket;
  delete tts;
  return 0;
}

#ifdef LINUX
static void sigpipe(int par)
{
    printf("sigpipe(%d)\n", par);
    fflush(stdout);
    return;
}

int exit_done = 0;
static void sigterm(int par)
{
    printf("got signal, terminating...\n");
    if(server) {
	ServerSocket*tmp = server;
	server = 0;
	delete tmp;
    } 
    if(capi) {
	SimpleCapi*tmp = capi;
	capi = 0;
	delete tmp;
    }
    if(log) {
	Logger*tmp = log;
	log = 0;
	delete tmp;
    }
    if(!exit_done)
    {
	exit_done = 1;
	exit(1);
    }
}
#endif

static void readline(FILE*fi, char*line, int maxlen) {
    int pos = 0;
    while(!feof(fi)) {
        if(!fread(&line[pos],1,1,fi))
            break;
        if(line[pos] == 13 || line[pos]==10)
            break;
	if(pos<maxlen-1)
	    pos++;
    }
    line[pos]=0;

    /* cut off comments */
    char*x=strchr(line,'#');
    if(x) {
        *x=0;
	pos = x-line;
    }

    /* cut off trailing whitespace */
    while(pos>=1 && (line[pos-1]==32 || line[pos-1]==9)) {
	pos--;
	line[pos]=0;
    }
}

char* simplecapi_property_msisdn=0;
char* simplecapi_property_allcalls=0;
char* simplecapi_property_controller=0;

static void readProperties(char*name)
{
    FILE*fi=0;
    char line[256];
    char* logfilename = "ivr.log";
    int fileloglevel = 5;
    int screenloglevel = 1;
    int socketloglevel = -1;
    char* logserver = 0;
    char* logport = 0;
    int properties_allcalls = 0;

    if(name)
	fi=fopen(name, "rb");
    while(fi && !feof(fi))
    {
	readline(fi, line, 256);
	char*equal = strchr(line, '=');
	if(equal)
	{
	    *equal = 0;
	    equal++;
	    if (!strcasecmp(line, "file.loglevel"))
                fileloglevel=atoi(equal);
            else if (!strcasecmp(line, "file.name"))
                logfilename=strdup(equal);
            else if (!strcasecmp(line, "screen.loglevel"))
                screenloglevel=atoi(equal);
            else if (!strcasecmp(line, "server.loglevel"))
                socketloglevel=atoi(equal);
            else if (!strcasecmp(line, "server.name"))
                logserver=strdup(equal);
            else if (!strcasecmp(line, "server.port"))
                logport=strdup(equal);
            else if (!strcasecmp(line, "local.msisdn"))
                simplecapi_property_msisdn=strdup(equal);
            else if (!strcasecmp(line, "local.controller"))
                simplecapi_property_controller=strdup(equal);
            else if (!strcasecmp(line, "local.allcalls"))
                simplecapi_property_allcalls=strdup(equal);
	    else if (!strcasecmp(line,"ivr.port")) {
		int p = atoi(equal);
		if(p>0) {
		    port = p;
		} else {
		    printf("Invalid port: %s\n", equal);
		    exit(1);
		}
	    } else if(!strcasecmp(line,"realspeak.dictionary")) {
		if(*equal)
		    realspeak_dictionary = strdup(equal);
	    }
	}
    }
    log = new Logger();
    log->initLog(logfilename, fileloglevel,
	         logserver, logport, socketloglevel,
		 screenloglevel);
}

int main(int argn, char*argv[])
{
  int t;
  char _version[80];
  char * version = _version + 2;
  FILE*fi;
  char*properties="<none>";
  sprintf(_version, "%s", VERSION);
  for(t=1;t<argn;t++) {
      if(!strcmp(argv[t], "-h")) {
	  printf("-h this help text\n");
	  printf("-v display version number\n");
	  exit(0);
      } else if(!strcmp(argv[t], "-V") ||
	        !strcmp(argv[t], "-v")) {
	  printf("capiserver version %s\n", version);
	  exit(0);
      } else if(!strcmp(argv[t], "-cvs")) {
	  printf("$Id: CapiServer.cpp,v 1.40 2002/11/05 20:02:52 kramm Exp $\n");
	  exit(0);
      } else {
	  printf("Unknown argument. Try -h\n");
	  exit(1);
      }
  }

  if (fi=fopen("ivr.properties","rb")) {
      fclose(fi);
      properties = "ivr.properties";
      readProperties(properties);
  } else if(fi=fopen("/etc/ivr.properties","rb")) {
      fclose(fi);
      properties = "/etc/ivr.properties";
      readProperties(properties);
  } else if(fi=fopen("./ivr.properties","rb")) {
      fclose(fi);
      properties = "./ivr.properties";
      readProperties(properties);
  } else readProperties(0);

  log->logf("<notice> CapiServer Version %s started.\n", version);
  log->logf("<notice> using property file %s.\n", properties);
  capi = new SimpleCapi(log);
  capi->setProperty("msisdn", simplecapi_property_msisdn);
  capi->setProperty("allcalls", simplecapi_property_allcalls);
  capi->setProperty("controller", simplecapi_property_controller);

#ifdef LINUX
  signal(SIGPIPE, sigpipe);
  signal(SIGINT , sigterm);
  signal(SIGQUIT, sigterm);
  signal(SIGTERM, sigterm);
#endif
  server = new ServerSocket();

  if(!server->initSocket(port))
  {
      log->logf("<fatal> Couldn't create listening socket\n");
      return 0;
  }
  log->logf("<notice> accepting connections on port %d\n", port);
  while(1)
  {
      ClientSocket*socket = server->acceptConnection();

      if(socket) {
	createThread(&thread, callHandler, socket);	
      }
  }
  if(server) {
      delete server;
      server = 0;
  }
  return 0;
}

